import { execSync } from 'child_process';
import os from 'os';
import fs from 'fs';
import * as path from 'path';
import Progress from '../util/Progress';
import Analyzer from './Analyzer';

type ClassMapEntity = {
	name: string,
	filePath: string,
	shortName?: string | null,
	extends: string | null,
	children: string[],
	parentsBranch: string[],
	mixins: string[],
	mixedIn: string[]
};

type ClassMap = {
	[key: string]: ClassMapEntity
};

class Sources {

	private static repoNames: string[]								= [];
	private static repoUrls: string[]									= [];
	private static repoSourcePaths: string[]					= [];
	private static repoSourceCopyPaths: string[]			= [];
	private static repoSparseCheckoutPaths: string[]	= [];
	private static repoCollectFromPaths: string[]			= [];
	private static sourcesPath: string								= `${os.homedir()}/.doczilla_js_docs`;

	public static sourcesCopyPath											= path.resolve(__dirname, '../JsSources');
	public static dzAppPath														= `${Sources.sourcesCopyPath}/DZApp.js`;
	public static classMapPath				    						= `${Sources.sourcesCopyPath}/ClassMap.json`;

	public static ClassMap: ClassMap = {};

	private static init(): void {
		// refresh
		Sources.sourcesPath = '';
		Sources.repoNames = [];
		Sources.repoUrls = [];
		Sources.repoSourcePaths = [];
		Sources.repoSourceCopyPaths = [];
		Sources.repoSparseCheckoutPaths = [];
		Sources.repoCollectFromPaths = [];

		const SourcesConfig = JSON.parse(fs.readFileSync(path.resolve(__dirname,'../sources.config.json'), { encoding: 'utf-8' }).toString());

		Sources.sourcesPath = `${os.homedir()}/${SourcesConfig.cloneTo}`;
		for(const repoInfo of SourcesConfig.repos) {
			Sources.repoNames.push(repoInfo.name);
			Sources.repoUrls.push(repoInfo.url);
			Sources.repoSourcePaths.push(`${Sources.sourcesPath}/${repoInfo.name}`);
			Sources.repoSourceCopyPaths.push(`${Sources.sourcesCopyPath}/${repoInfo.name}`);
			Sources.repoSparseCheckoutPaths.push(repoInfo.sparseCheckout);
			Sources.repoCollectFromPaths.push(repoInfo.collectFrom);
		}
	}

	public static get(callback?: () => any): ClassMap {
		if(fs.existsSync(Sources.classMapPath)) {
			Sources.ClassMap = JSON.parse(fs.readFileSync(Sources.classMapPath).toString());
			Sources.repoNames = JSON.parse(fs.readFileSync(path.resolve(__dirname,'../sources.config.json'), { encoding: 'utf-8' }).toString()).repos.map((repo: { name: string; }) => repo.name);
			callback && callback();
		} else {
			Sources.update(callback);
		}
		return Sources.ClassMap;
	}

	public static getShortenedClassMap(): { [key: string]: { shortName: string | null, filePath: string, name: string } } {
		const classList: { [key: string]: { shortName: string | null, filePath: string, name: string } } = {};
		
		Object.keys(Sources.ClassMap).map((cls) => { 
			const scls = Sources.ClassMap[cls];
			classList[cls] = { name: cls, shortName: scls.shortName || null, filePath: scls.filePath };
		});

		return classList;
	}

	public static getRepoNames(): string[] {
		return Sources.repoNames;
	}

	public static time(): Date {
		return fs.statSync(Sources.classMapPath).birthtime;
	}

	public static findClass(className: string): ClassMapEntity | null {
		if(Sources.ClassMap[className])
			return Sources.ClassMap[className];
		const keys = Object.keys(Sources.ClassMap).filter((key) => {
			return Sources.ClassMap[key].shortName === className;
		});
		if(keys.length === 1)
			return Sources.ClassMap[keys[0]];
		return null;
	}

	public static update(callback?: () => any): void {
		Sources.init();

		Sources.clone();
		Sources.copy();
		Sources.collect();

		Analyzer.analyze(Sources.ClassMap, callback);
	}

	private static clone(): void {
		if(!fs.existsSync(Sources.sourcesPath))
			fs.mkdirSync(Sources.sourcesPath);

		const progress = Progress.start('Sources update, Step 1 [Clone/Pull]:', Sources.repoNames.length);
		for(let i = 0; i < Sources.repoNames.length; i++) {
			if(fs.existsSync(Sources.repoSourcePaths[i])) {
				execSync('git pull', { encoding: 'utf8', cwd: Sources.repoSourcePaths[i], stdio: [] });
				progress.next();
			} else {
				execSync(Sources.getCloneCommand(Sources.repoUrls[i]), { encoding: 'utf8', cwd: Sources.sourcesPath, stdio: [] });
				execSync(`git sparse-checkout set ${Sources.repoSparseCheckoutPaths[i]}`, { encoding: 'utf8', cwd: Sources.repoSourcePaths[i], stdio: [] });
				progress.next();
			}
		}

		progress.finish();
	}

	private static copy(): void {
		if(fs.existsSync(Sources.sourcesCopyPath))
			fs.rmSync(Sources.sourcesCopyPath, { recursive: true, force: true });
		fs.mkdirSync(Sources.sourcesCopyPath);

		const progress = Progress.start('Sources update, Step 2 [Copy]:', Sources.repoNames.length);

		for(let i = 0; i < Sources.repoNames.length; i++) {
			fs.mkdirSync(Sources.repoSourceCopyPaths[i]);
			fs.cpSync(`${Sources.repoSourcePaths[i]}/${Sources.repoCollectFromPaths[i]}`, Sources.repoSourceCopyPaths[i], { recursive: true });
			progress.next();
		}

		progress.finish();
	}

	private static collect(): void {
		if(fs.existsSync(Sources.dzAppPath))
			fs.rmSync(Sources.dzAppPath, { force: true });

		Sources.ClassMap = {};

		const buildorders = [];
		let totalBuildordersLength = 0;

		for(let i = 0; i < Sources.repoNames.length; i++) {
			const bo = fs.readFileSync(`${Sources.repoSourceCopyPaths[i]}/.buildorder`, 'utf8').toString().replaceAll('\r', '').split('\n').filter((str) => { return str.length > 0; });
			buildorders.push(bo);
			totalBuildordersLength += bo.length;
		}

		const progress = Progress.start('Sources update, Step 3 [Collect]:', totalBuildordersLength);

		for(let i = 0; i < Sources.repoNames.length; i++) {
			Sources.processBuildorder(buildorders[i], Sources.repoSourceCopyPaths[i], Sources.repoNames[i], progress);
		}

		progress.finish();
	}

	private static processBuildorder(buildorder: string[], sourcesRoot: string, pathPrefix: string, progress: Progress): void {
		const z8DefinePattern: RegExp = /Z8\.define\('([^']+)'/g;

		for(const filePath of buildorder) {
			const content = fs.readFileSync(`${sourcesRoot}/${filePath}`, 'utf8').toString();
			const defines: string[] = Array.from(content.matchAll(z8DefinePattern), match => match[1]);
			for(const define of defines) {
				Sources.ClassMap[define] = { name: define, filePath: `${pathPrefix}/${filePath}`, children: [], extends: null, parentsBranch: [], mixins: [], mixedIn: [] };
			}
			fs.writeFileSync(Sources.dzAppPath, content + '\n', { flag: 'a+', encoding: 'utf8' });
			progress.next();
		}
	}

	private static getCloneCommand(repoUrl: string): string {
		return `git clone --depth=1 --single-branch --branch dev --filter=blob:none --sparse ${repoUrl}`;
	}
}

export { Sources, ClassMap, ClassMapEntity };
